'''
Computations of dynamics (i.e., reward and transition functions) when 
increasing the frequency of samples from the top k most similar people.

Author: Nele Albers
Date: April 2022
'''


import copy
import itertools
import numpy as np

import Utils as util


def compute_dynamics_p(data_p, abstract_states, num_feat, 
                       num_act, feat_sel):
    """Compute estimated reward and transition function from samples.

    Args:
        data_p (list): <s0, s1, a, r>-samples.
        abstract_states (list): Abstract states.
        num_feat (int): Number of features used.
        num_act (int): Number of actions.
        feat_sel (list): Selected features

    Returns:
        np-array: reward function.
        np-array: transition function.

    """
    
    num_samples_p = len(data_p)
    
    trans_func = np.zeros((int(2 ** num_feat), num_act, int(2 ** num_feat)))
    reward_func = np.zeros((int(2 ** num_feat), num_act, int(2 ** num_feat)))
    reward_func_count = np.zeros((int(2 ** num_feat), num_act, int(2 ** num_feat)))
    
    for s_ind, s in enumerate(abstract_states):
        for data_index in range(num_samples_p):  # for each data sample
            if list(np.take(np.array(data_p[data_index][0]), feat_sel)) == s:
                
                a = data_p[data_index][2]  # action
                r = data_p[data_index][3]
                next_state_index = abstract_states.index(list(np.take(data_p[data_index][1], feat_sel)))
                
                trans_func[s_ind, a, next_state_index] += 1
                reward_func[s_ind, a, next_state_index] += r
                reward_func_count[s_ind, a, next_state_index] += 1
       
        # Normalize reward and transition functions
        for a in range(num_act):
            summed = sum(trans_func[s_ind, a])
            if summed > 0:
                trans_func[s_ind, a] /= summed
            # If we have no data on a state-action combination, 
            # we assume equal probability of transitioning to each other state
            else:
                trans_func[s_ind, a] = np.ones(int(2 ** num_feat)) / (2 ** num_feat)
            
            for s_prime_ind in range(len(abstract_states)):
                if reward_func_count[s_ind, a, s_prime_ind] > 0:
                    reward_func[s_ind, a, s_prime_ind] /= reward_func_count[s_ind, a, s_prime_ind]
        
    return reward_func, trans_func


def compute_dynamics(data_in, effort_mean, feat_sel, user_ids,
                     user_ids_comp_policy, traits,
                     traits_ids, increase_factor,
                     cutoff_fraction,
                     num_act=5, discount_factor = 0.85):
    """Compute the optimal policy for level 4 of algorithm complexity.

    Args:
        data_in (list): List with samples of the form <s0, s1, a, r>.
        effort_mean (float): Mean value of effort responses.
        feat_sel (list): Selected features.
        user_ids (list): IDs of users corresponding to samples in data.
        user_ids_comp_policy (list): IDs of users for whom we need to compute policies.
        traits (np-array): Traits for each person to use as basis for similarity.
        traits_ids (list): IDs corresponding to traits.
        increase_factor (int): Number of times to copy a sample from a more similar person.
        cutoff_fraction (float): Ratio of people to consider as most similar, [0, 1].
        num_act (int): Number of possible actions.
        discount_factor (float): Discount factor for Q-value computation

    Returns:
        dict: similarity-weighted reward function per person (ID)
        dict: similarity-weighted transition function per person (ID)

    """
    
    data = copy.deepcopy(data_in)
    
    num_assigned = len(user_ids_comp_policy)  # number of people for whom we need to calculate policies
    num_feat = len(feat_sel)  # number of selected features that we now consider
    num_samples = len(data)
    
    # All effort responses
    list_of_efforts = list(np.array(data, dtype=object)[:, 3].astype(int))
    
    # Map effort responses to rewards from -1 to 1, with the mean mapped to 0.
    map_to_rewards = util.get_map_effort_reward(effort_mean, 
                                                output_lower_bound = -1, 
                                                output_upper_bound = 1, 
                                                input_lower_bound = 0, 
                                                input_upper_bound = 10)
    reward_list = util.map_efforts_to_rewards(list_of_efforts, map_to_rewards)
    # now write these obtained reward values into "data"
    for i in range(len(reward_list)):
        data[i][3] = reward_list[i]
    
    reward_funcs = {}
    trans_funcs = {}

    # For each person for whom we need to calculate a policy.
    for p1 in range(num_assigned):
        
        # Get index of traits for this person
        trait_index_p1 = traits_ids.index(user_ids_comp_policy[p1])
        
        data_p = copy.deepcopy(data)
        
        
        # Get Euclidean distance of traits to all other people
        traits_distances = []
        unique_user_ids = list(set(user_ids))
        for unique_user_idx in range(len(unique_user_ids)):
            trait_index_unique_user = traits_ids.index(unique_user_ids[unique_user_idx])
            traits_distances.append(np.linalg.norm(traits[trait_index_p1] - traits[trait_index_unique_user]))
        
        
        traits_distances_sorted_idx = np.argsort(traits_distances)
        cutoff = int(cutoff_fraction * len(unique_user_ids))
        unique_user_ids_cutoff = [unique_user_ids[i] for i in range(len(unique_user_ids)) if i in traits_distances_sorted_idx[0:cutoff]]
        
        
        # Go through all samples
        # We will increase their frequency if they belong to the most
        # similar people
        for p2 in range(num_samples):
            
            # Increase frequency of sample
            if user_ids[p2] in unique_user_ids_cutoff:
                for inc in range(increase_factor):
                    data_p.append(data[p2])
            
        
        abstract_states = [list(i) for i in itertools.product([0, 1], repeat = num_feat)]
    
        # Estimate reward and transition functions
        reward_func, trans_func = compute_dynamics_p(data_p, 
                                                     abstract_states,
                                                     num_feat, 
                                                     num_act, 
                                                     feat_sel)
        
        # Save reward and transition function for this person
        reward_funcs[user_ids_comp_policy[p1]] = reward_func
        trans_funcs[user_ids_comp_policy[p1]] = trans_func
        
    return reward_funcs, trans_funcs
