'''
Common functions for RL-related computations.
'''


from copy import deepcopy
import numpy as np


def get_map_effort_reward(effort_mean, output_lower_bound, 
                          output_upper_bound = 1, input_lower_bound = 0, 
                          input_upper_bound = 10):
    """
    Computes a mapping from effort responses to rewards.
    
    Args:
        effort_mean (float): mean effort response, to be mapped to halfway 
                             between output_lower_bound and output_upper_bound
        output_lower_bound (float): lowest value on output scale
        output_upper_bound (float, default 1): highest value on output scale
        input_lower_bound (int, default 0): lowest value on input scale, to 
                                            be mapped to output_lower_bound
        input_upper_bound (int, default 10): highest value on input scale, 
                                             to be mapped to output_upper_bound
                                             
    Returns:
        dictionary: maps effort responses (integers) to output scale values (float)
    """
    map_to_rew = {}
    
    # We can already map the endpoints of the input scale to the output scale
    map_to_rew[input_lower_bound] = output_lower_bound
    map_to_rew[input_upper_bound] = output_upper_bound
    
    # The mean value on the output scale
    mean_output = (output_upper_bound - output_lower_bound)/2 + output_lower_bound
    output_length_half = mean_output - output_lower_bound
    
    input_length_lower_half = effort_mean - input_lower_bound
    input_length_upper_half = input_upper_bound - effort_mean
    
    inc_lower_half = output_length_half / input_length_lower_half
    inc_upper_half = output_length_half / input_length_upper_half
    
    # Compute output scale values for input scale below mean.
    idx = 1
    for i in range(input_lower_bound + 1, int(np.ceil(effort_mean))):
        map_to_rew[i] = output_lower_bound + idx * inc_lower_half
        idx += 1
    
    # Compute output scale values for input scale above mean.
    idx = 1
    for i in range(input_upper_bound - 1, int(np.floor(effort_mean)), -1):
        map_to_rew[i] = output_upper_bound - idx * inc_upper_half
        idx += 1
    
    if np.floor(effort_mean) == effort_mean:
        map_to_rew[effort_mean] = mean_output
    
    # Need to check whether effort_mean is an integer that we need to map
    # to mean_output on the output scale.
    return map_to_rew


def map_efforts_to_rewards(list_of_efforts, map_to_rewards):
    """
    Maps effort responses to rewards.
    
    Args:
        list_of_efforts (list of int): effort responses to be mapped
        map_to_rewards (dictionary): maps effort responses to reward values (float)
    
    Returns:
        list (float): resulting reward values
    """
    rewards = []
    for e in list_of_efforts:
        rewards.append(map_to_rewards[e])
    
    return rewards
	
	
def policy_evaluation(observation_space_size, discount_factor, trans_func,
                      reward_func, policy, q_vals, 
                      update_tolerance = 0.000000001,
                      reward_dep_next_state = False):
    """
    Returns the q-values for a specific policy.
    
    Args:
        observation_space_size: number of observations
        discount_factor: discount factor of MDP
        trans_func: transition function
        reward_func: reward function
        policy: policy to compute q-values for
        update_tolerance: precision
        reward_dep_next_state (boolean, default: False): whether reward depends also on the next state
    """
    q_vals_new = deepcopy(q_vals)
    num_act = len(trans_func[0])
    
    update = 1 # max. change in Q-value in one iteration
   
    if not reward_dep_next_state: # reward does not also depend on next state
        while update > update_tolerance:
            update = 0
            for s in range(observation_space_size):
                for a in range(num_act):
                    q_vals_new[s, a] = reward_func[s, a] + discount_factor * sum([trans_func[s, a, s_prime] * q_vals[s_prime, int(policy[s_prime])] for s_prime in range(observation_space_size)])
                    update = max(update, abs(q_vals[s, a] - q_vals_new[s, a]))
            q_vals = deepcopy(q_vals_new)
    
    else: # reward does also depend on the next state
        while update > update_tolerance:
            update = 0
            for s in range(observation_space_size):
               
                for a in range(num_act):
                    q_vals_new[s, a] = sum([trans_func[s, a, s_prime] * (discount_factor * q_vals[s_prime, int(policy[s_prime])] + reward_func[s, a, s_prime]) for s_prime in range(observation_space_size)])
                    
                    update = max(update, abs(q_vals[s, a] - q_vals_new[s, a]))
            q_vals = deepcopy(q_vals_new)
        
    return q_vals_new
    

def get_Q_values_opt_policy(discount_factor, trans_func, reward_func,
                            reward_dep_next_state = False):
    """
    Returns the Q-values for each state under the optimal policy as well
    as the optimal policy.
    
    Args:
        discount_factor: discount factor of MDP
        trans_func: transition function (dim.: num_states x num_actions x num_states)
        reward_func: reward function (dim.: num_states x num_actions)
        reward_dep_next_state (boolean, default: False): whether reward depends also on the next state
    """
    min_iterations = 100
    num_states = len(trans_func)
    num_act = len(trans_func[0])
    q_vals = np.zeros((num_states, num_act))

    policy = np.zeros(num_states)
    policy_new = np.ones(num_states)
    it = 0
    
    while (not np.array_equal(policy, policy_new)) or it < min_iterations:
        q_vals = policy_evaluation(num_states, discount_factor, 
                                   trans_func, reward_func, policy, q_vals,
                                   reward_dep_next_state = reward_dep_next_state)
        policy = policy_new
        policy_new = np.array([np.argmax(q_vals[s]) for s in range(num_states)])
        it += 1
    
    return q_vals, policy_new
