"""
Extract RL samples from the conversational sessions data from the "Data"-folder.
Overall we get 2366 samples of the form <s, a, r, s'>.

Author: Nele Albers
Date: February 2023

Required files:
    - Data/conversational_sessions_anonym.xlsx
    - Data/pre_questionnaire_anonym.csv
    - Data/prescreening_prolific_numeric_anonym.csv
    - Data/post_questionnaire_anonym.csv
Output files:
    - extract_RL_samples_output.txt
    - Analysis/Data/rl_samples_list_binary_activities.csv
    - Analysis/Data/all_states.csv
"""


import numpy as np
import os
import pandas as pd


# Number of state features: we also measured two other features besides the
# 8 that we used in our analysis. These additional features were about people's
# physical activity identity and smoker identity. Notably, the corresponding
# questions are not part of the COM-B self-evaluation questionnaire. Instead,
# we had included the measurement of these two features for exploratory purposes.
# As one's identity is unlikely to change as a result of a single 
# persuasive attempt, however, we decided to incorporate
# these two types of identities as user characteristics rather than state
# features. This was possible, as we had used more elaborate questions to 
# assess these two types of identities in the pre-questionnaire. For the sake
# of completeness, we still provide the data on these two state
# features. Yet, we do not consider them in our selection of state features
# in our later analysis.
NUM_STATE_FEATURES = 10

# Names of the data columns corresponding to the 5 states
STATES_NAMES_LIST = ["state_0", "state_1", "state_2", "state_3", "state_4"]


def cronbach_alpha(df):
    """
    Calculate cronbach's alpha.
    Code from https://towardsdatascience.com/cronbachs-alpha-theory-and-application-in-python-d2915dd63586.
    """

    # 1. Transform the df into a correlation matrix
    df_corr = df.corr()

    # 2.1 Calculate N
    # The number of variables equals the number of columns in the df
    N = df.shape[1]

    # 2.2 Calculate R
    # For this, we'll loop through the columns and append every
    # relevant correlation to an array calles "r_s". Then, we'll
    # calculate the mean of "r_s"
    rs = np.array([])
    for i, col in enumerate(df_corr.columns):
        sum_ = df_corr[col][i+1:].values
        rs = np.append(sum_, rs)
    mean_r = np.mean(rs)

    # 3. Use the formula to calculate Cronbach's Alpha 
    cronbach_alpha = (N * mean_r) / (1 + (N - 1) * mean_r)

    return cronbach_alpha


def get_similarity_variables_pre_questionnaire(df, output_file_name):
    """
    Get variables from the pre-questionnaire.
    """
    df_preq_q = pd.read_csv(curr_directory + "/Data/pre_questionnaire_anonym.csv")


    # We only provide data for people who passed enough attention checks
    # and gave informed consent.
    df_preq_q = df_preq_q.drop(columns = ["Finished",
                                          "Consent", 
                                          "Check1", 
                                          "Check2", 
                                          "Check3", 
                                          "Check4", 
                                          "Check5"])

    df = pd.merge(df, df_preq_q, on = "rand_id")


    # Compute indices
    # Personality items were measured on a scale from 1 to 7
    df['Pers_Extraversion'] = 0.5 * (df['Personality_1']- (df['Personality_6'] - 8)) 
    df['Pers_Agreeableness'] = 0.5 * (df['Personality_7'] - (df['Personality_2'] - 8)) 
    df['Pers_Conscientiousness'] = 0.5 * (df['Personality_3'] - (df['Personality_8'] - 8)) 
    df['Pers_ES'] = 0.5 * (df['Personality_9'] - (df['Personality_4'] - 8)) 
    df['Pers_OE'] = 0.5 * (df['Personality_5'] - (df['Personality_10'] - 8)) 
    df["PA_Identity"] = 1/9 * (df["PA_Identity_1"] + df["PA_Identity_2"] + df["PA_Identity_3"] + df["PA_Identity_4"] + df["PA_Identity_5"] + df["PA_Identity_6"] + df["PA_Identity_8"] + df["PA_Identity_9"] + df["PA_Identity_7"])  
    df["Quitting_Self_Identity"] = 1/3 * (df["Quitting_Self_Identity_1"] + df["Quitting_Self_Identity_2"] + df["Quitting_Self_Identity_3"]) 
    df["Smoker_Self_Identity"] = 1/3 * (df["Smoker_Self_Identity_1"] + df["Smoker_Self_Identity_2"] + df["Smoker_Self_Identity_3"]) 
    df["Nonsmoker_Self_Identity"] = 1/3 * (df["Nonsmoker_Self_Identity_1"] + df["Nonsmoker_Self_Identity_2"] + df["Nonsmoker_Self_Identity_3"]) 
    df["Smoker_Group_Identity"] = 1/5 * (df["Smoker_Group_Identity_1"] + df["Smoker_Group_Identity_2"] - df["Smoker_Group_Identity_3"] + 6 - df["Smoker_Group_Identity_4"] + 6 + df["Smoker_Group_Identity_5"])
    df["Quitting_Self_Efficacy"] = 1/4 * (df["Quitting_Self_Efficacy_1"] + df["Quitting_Self_Efficacy_2"] + df["Quitting_Self_Efficacy_3"] + df["Quitting_Self_Efficacy_4"]) 
    df["Nonsmoker_Group_Identity"] = 1/5 * (df["Nonsmoker_Group_Identity_1"] + df["Nonsmoker_Group_Identity_2"] - df["Nonsmoker_Group_Identity_3"] + 6 - df["Nonsmoker_Group_Identity_4"] + 6 + df["Nonsmoker_Group_Identity_5"])
    # NFC was measured on a scale from 1 to 5, items 2 and 3 are scored
    # in reverse
    df["NFC"] = 1/3 * (df["NFC_1"] - (df["NFC_2"] - 6) - (df["NFC_3"] - 6))


    # Compute Cronbach's alpha values for indices
    df_sub = df[["PA_Identity_1", "PA_Identity_2", "PA_Identity_3", "PA_Identity_4",
                 "PA_Identity_5", "PA_Identity_6", "PA_Identity_7", "PA_Identity_8",
                 "PA_Identity_9"]]
    print(len(df_sub))
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha physical activity identity:", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'w') as f:
        f.write("Cronbach's alpha physical activity identity: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")

    df_sub = df[["Quitting_Self_Identity_1", "Quitting_Self_Identity_2", 
                 "Quitting_Self_Identity_3"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha quitter self-identity:", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha quitter self-identity: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")

    df_sub = df[["Smoker_Self_Identity_1", "Smoker_Self_Identity_2", 
                 "Smoker_Self_Identity_3"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha smoker self-identity:", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha smoker self-identity: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")

    df_sub = df[["Nonsmoker_Self_Identity_1", "Nonsmoker_Self_Identity_2", 
                 "Nonsmoker_Self_Identity_3"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha nonsmoker self-identity:", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha nonsmoker self-identity: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")

    df["Smoker_Group_Identity_3"] = - df["Smoker_Group_Identity_3"] + 6  # scored in reverse
    df["Smoker_Group_Identity_4"] = - df["Smoker_Group_Identity_4"] + 6  # scored in reverse
    df_sub = df[["Smoker_Group_Identity_1", "Smoker_Group_Identity_2", 
                 "Smoker_Group_Identity_3", "Smoker_Group_Identity_4",
                 "Smoker_Group_Identity_5"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha smoker group-identity:", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha smoker group-identity: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")

    df_sub = df[["Quitting_Self_Efficacy_1", "Quitting_Self_Efficacy_2", 
                 "Quitting_Self_Efficacy_3", "Quitting_Self_Efficacy_4"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha quitting self-efficacy:", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha quitting self-efficacy: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")

    df["Nonsmoker_Group_Identity_3"] = - df["Nonsmoker_Group_Identity_3"] + 6  # scored in reverse
    df["Nonsmoker_Group_Identity_4"] = - df["Nonsmoker_Group_Identity_4"] + 6  # scored in reverse
    df_sub = df[["Nonsmoker_Group_Identity_1", "Nonsmoker_Group_Identity_2", 
                 "Nonsmoker_Group_Identity_3", "Nonsmoker_Group_Identity_4",
                 "Nonsmoker_Group_Identity_5"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha nonsmoker group-identity: ", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha nonsmoker group-identity: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")

    df["NFC_2"] = - df["NFC_2"] + 6  # scored in reverse
    df["NFC_3"] = - df["NFC_3"] + 6  # scored in reverse
    df_sub = df[["NFC_1", "NFC_2", "NFC_3"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha NFC:", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha NFC: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")


    # Scale all of the variables to the interval [0, 1]
    df["TTM_PA"] = (df["TTM_PA"] - 1)/(5 - 1)
    df["Pers_Extraversion"] = (df["Pers_Extraversion"] - 1)/(7 - 1)
    df["Pers_Agreeableness"] = (df["Pers_Agreeableness"] - 1)/(7 - 1)
    df["Pers_Conscientiousness"] = (df["Pers_Conscientiousness"] - 1)/(7 - 1)
    df["Pers_ES"] = (df["Pers_ES"] - 1)/(7 - 1)
    df["Pers_OE"] = (df["Pers_OE"] - 1)/(7 - 1)
    df["PA_Identity"] = (df["PA_Identity"] - 1)/(5 - 1)
    df["Quitting_Self_Identity"] = (df["Quitting_Self_Identity"] - 1)/(5 - 1)
    df["Smoker_Self_Identity"] = (df["Smoker_Self_Identity"] - 1)/(5 - 1)
    df["Nonsmoker_Self_Identity"] = (df["Nonsmoker_Self_Identity"] - 1)/(5 - 1)
    df["Smoker_Group_Identity"] = (df["Smoker_Group_Identity"] - 1)/(5 - 1)
    df["Quitting_Self_Efficacy"] = (df["Quitting_Self_Efficacy"] - 1)/(5 - 1)
    df["Nonsmoker_Group_Identity"] = (df["Nonsmoker_Group_Identity"] - 1)/(5 - 1)
    df["NFC"] = (df["NFC"] - 1)/(5 - 1)  # 5-point Likert scale
    df["Quit_Before_24h"] = df["Quit_Before_24h"] - 1  # binary variable with values 1, 2
    df["Num_Quit_24h_Last_Year"] = (df["Num_Quit_24h_Last_Year"] - 1)/(4 - 1)  # values go from 1 to 4
    df["Time_Since_Last_24h_Quit_Attempt"] = (df["Time_Since_Last_24h_Quit_Attempt"] - 1)/(5 - 1)  # values go from 1 to 5
    df["Sitting_Hours_Weekday"] = df["Sitting_Hours_Weekday"] / 24  # values go from 0 to 24
    df["Sitting_Hours_Weekendday"] = df["Sitting_Hours_Weekendday"] / 24  # values go from 0 to 24


    # Drop columns used for computing indices
    df = df.drop(columns = ["Personality_1", "Personality_2",
                            "Personality_3", "Personality_4", "Personality_5",
                            "Personality_6", "Personality_7", "Personality_8", 
                            "Personality_9", "Personality_10",
                            "PA_Identity_1", "PA_Identity_2",
                            "PA_Identity_3", "PA_Identity_4", "PA_Identity_5",
                            "PA_Identity_6", "PA_Identity_7", "PA_Identity_8",
                            "PA_Identity_9", 
                            "NFC_1", "NFC_2", "NFC_3",
                            "Nonsmoker_Group_Identity_1", 
                            "Nonsmoker_Group_Identity_2", 
                            "Nonsmoker_Group_Identity_3", 
                            "Nonsmoker_Group_Identity_4", 
                            "Nonsmoker_Group_Identity_5", 
                            "Quitting_Self_Efficacy_1", 
                            "Quitting_Self_Efficacy_2", 
                            "Quitting_Self_Efficacy_3", 
                            "Quitting_Self_Efficacy_4", 
                            "Nonsmoker_Self_Identity_1",
                            "Nonsmoker_Self_Identity_2",
                            "Nonsmoker_Self_Identity_3",
                            "Smoker_Self_Identity_1",
                            "Smoker_Self_Identity_2",
                            "Smoker_Self_Identity_3",
                            "Quitting_Self_Identity_1",
                            "Quitting_Self_Identity_2",
                            "Quitting_Self_Identity_3",
                            "Smoker_Group_Identity_1",
                            "Smoker_Group_Identity_2",
                            "Smoker_Group_Identity_3",
                            "Smoker_Group_Identity_4",
                            "Smoker_Group_Identity_5"])

    return df


def get_inv_post_questionnaire(df, output_file_name):
    """
    Get involvement data from the post-questionnaire.
    """

    df_post_q = pd.read_csv(curr_directory + "/Data/post_questionnaire_anonym.csv")
    df = pd.merge(df, df_post_q, on = 'rand_id', how = 'left')

    # Only included data on people passing enough attention checks
    df = df.drop(columns = ["Check1", "Check2", "Finished"])

    # Compute index as the mean of the 3 involvement items
    df["Involvement"] = (df["Activity_Involvement_1"] + df["Activity_Involvement_2"] + df["Activity_Involvement_3"])/3

    # Compute Cronbach's alpha
    df_sub = df[['Activity_Involvement_1',
                 "Activity_Involvement_2", 
                 'Activity_Involvement_3']]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha involvement:", round(alpha, 2), "(N=", str(len(df_sub)) + ")")
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha involvement: " + str(round(alpha, 2)) + " (N=" + str(len(df_sub)) + ")\n")

    # Do not need single involvement items anymore
    df = df.drop(columns = ['Activity_Involvement_1',
                            "Activity_Involvement_2", 
                            'Activity_Involvement_3'])

    return df


def get_char_prescreening_questionnaire(df):
    """
    Get the TTM-stage for quitting smoking from the prescreening questionnaire.
    """
    df_prescreening_q = pd.read_csv(curr_directory + "/Data/prescreening_questionnaire_anonym.csv")
    # Only need column on wanting to quit smoking and the ID
    df_prescreening_q = df_prescreening_q[["rand_id", "Quit_Smoking"]]

    df = pd.merge(df, df_prescreening_q, on = "rand_id", how = 'left')

    # People are only in the preparation stage if they also had at least one quit
    # attempt of at least 24h in the last year.
    # 0 stands for preparation stage, 1 for contemplation stage
    df["TTM_Smoking"] = [0 if df.iloc[i]["Quit_Smoking"] == 1 and df.iloc[i]["Num_Quit_24h_Last_Year"] > 0 else 1 for i in range(len(df))]

    # No longer need column on quitting smoking
    df = df.drop(columns = ["Quit_Smoking"])

    return df


def get_char_prolific_profile(df, output_file_name):
    """
    Get variables captured from Prolific profiles.
    """

    df_prol_prof = pd.read_csv(curr_directory + "/Data/prescreening_prolific_numeric_anonym.csv")
    df = pd.merge(df, df_prol_prof, on = "rand_id")

    df = df.drop(columns=["Num_Previous_Submissions", 
                          "prolific_score"])

    # From the categorical gender variable with male, female, other, create
    # two new binary variables for male and female gender
    df["Gender_Male"] = (df["Gender identity"] == 1).astype(int)
    df["Gender_Female"] = (df["Gender identity"] == 2).astype(int)

    # Map potential similarity variables to interval [0, 1]
    df["Highest education level completed"] = (df["Highest education level completed"] - 1)/(7 - 1)
    df["Household Size"] = (df["Household Size"] - 1)/(11 - 1)
    df["Smoking frequency"] = (df["Smoking frequency"] - 1)/(5 - 1)
    df["Weekly Exercise"] = (df["Weekly Exercise"] - 1)/(3-1)
    df["Household Income (GBP)"] = (df["Household Income (GBP)"] - 1)/(13 - 1)
    df["Personal Income (GBP)"] = (df["Personal Income (GBP)"] - 1)/(12 - 1)
    df["Quit smoking"] = (df["Quit smoking"] - 1)/(6 - 1)
    df["Smoking status"] = (df["Smoking status"] - 1)/(4 - 1)
    df["Socioeconomic Status"] = (df["Socioeconomic Status"] - 1)/(10 - 1)
    min_age = df["age_bin"].min()
    max_age = df["age_bin"].max()
    df["age_bin"] = (df["age_bin"] - min_age)/(max_age - min_age)

    print("Minimum age bin:", min_age, "Maximum age bin:", max_age)
    with open(output_file_name, 'a') as f:
        f.write("\nMinimum age bin: " + str(min_age) + "\n")
        f.write("Maximum age bin: " + str(max_age) + "\n\n")


    # Rename some columns
    df = df.rename(columns={"Highest education level completed": "Highest_Completed_Education_Level",
                           "Household Size": "Household_Size", 
                           "Smoking frequency": "Smoking_Frequency",
                           "Weekly Exercise": "Weekly_Exercise",
                           "Household Income (GBP)": "Household_Income",
                           "Personal Income (GBP)": "Personal_Income",
                           "Quit smoking": "Quit_Smoking_Status",
                           "Smoking status": "Smoking_Status",
                           "Socioeconomic Status": "Socioeconomic_Status"})
    
    return df


def save_transition_single(df, pers_single, p, s, state_list_0, rew_list,
                           action_type_list):
    """
    Save one transition to dataframe.

    Parameters
    ----------
    df : dataframe
        Complete source dataframe
    pers_single : list
        dataframe for transitions
    p : int
        Person index (i.e., row in df)
    s : int
        First state index (i.e., 0-3)
    state_list_0 : string
        First state
    rew_list : list
        List of efforts for this person p
    action_type_list : list
        List of action types for this person p

    Returns
    -------
    pers_single : list
        pers_single with transition added.

    """
    state_list_1 = str(df.iloc[p][STATES_NAMES_LIST[s + 1]]).split('|')

    # We might not have info on all features for the next state
    for missing_feat in range(len(state_list_1), NUM_STATE_FEATURES):
        state_list_1.append(np.nan)

    pers_single.append([list(map(float, state_list_0)),
                        list(map(float, state_list_1)),
                        int(action_type_list[s]),
                        int(rew_list[s]),
                        df.iloc[p]["rand_id"],
                        df.iloc[p]["Pers_OE"],
                        df.iloc[p]["Pers_Extraversion"],
                        df.iloc[p]["Pers_ES"],
                        df.iloc[p]["Pers_Conscientiousness"],
                        df.iloc[p]["Pers_Agreeableness"],
                        df.iloc[p]["TTM_PA"],
                        df.iloc[p]["PA_Identity"],
                        df.iloc[p]["Involvement"],
                        s,
                        df.iloc[p]["Quitting_Self_Identity"],
                        df.iloc[p]["Smoker_Self_Identity"],
                        df.iloc[p]["Nonsmoker_Self_Identity"],
                        df.iloc[p]["Smoker_Group_Identity"],
                        df.iloc[p]["Nonsmoker_Group_Identity"],
                        df.iloc[p]["Quitting_Self_Efficacy"],
                        df.iloc[p]["NFC"],
                        df.iloc[p]["Quit_Before_24h"],
                        df.iloc[p]["Num_Quit_24h_Last_Year"],
                        df.iloc[p]["Time_Since_Last_24h_Quit_Attempt"],
                        df.iloc[p]["Sitting_Hours_Weekday"],
                        df.iloc[p]["Sitting_Hours_Weekendday"],
                        df.iloc[p]["age_bin"],
                        df.iloc[p]["Highest_Completed_Education_Level"],
                        df.iloc[p]["Household_Size"],
                        df.iloc[p]["Smoking_Frequency"],
                        df.iloc[p]["Weekly_Exercise"],
                        df.iloc[p]["Household_Income"],
                        df.iloc[p]["Personal_Income"],
                        df.iloc[p]["Quit_Smoking_Status"],
                        df.iloc[p]["Smoking_Status"],
                        df.iloc[p]["Socioeconomic_Status"],
                        df.iloc[p]["Gender_Male"],
                        df.iloc[p]["Gender_Female"],
                        df.iloc[p]["Gender identity"],
                        df.iloc[p]["TTM_Smoking"]
                        ])
    
    return pers_single


def save_data_person(df, p, s1_failed, s2_failed, s3_failed,
                     all_states, all_states_ids, all_states_sessions,
                     pers_single):
    """
    Save data for single person.

    Parameters
    ----------
    df : dataframe
        Complete source dataframe.
    p : int
        Person index (i.e., row in df)
    s1_failed : list
        IDs of people who failed session 1
    s2_failed : list
        IDs of people who failed session 2
    s3_failed : list
        IDs of people who failed session 3
    all_states : list
        List of all states
    all_states_ids : list
        List of IDs corresponding to states in all_states
    all_states_sessions : list
        Session IDs corresponding to states in all_states
    pers_single : list
        Dataframe with transitions

    Returns
    -------
    all_states : list
        all_states plus newly added states
    all_states_ids : list
        all_states_ids plus newly added ids
    all_states_sessions : list
        all_states_sessions plus newly added session IDs
    pers_single : list
        pers_single plus newly added transitions

    """
    
    # Get efforts
    rew_list = str(df.iloc[p]["effort_list"])
    if len(rew_list) > 1:
        rew_list = rew_list.split('|')
    # Single effort for this person
    else:
        rew_list = [rew_list]

    # Get corresponding action types (i.e., persuasive strategies)
    action_type_list = str(df.iloc[p]["action_type_index_list"])
    if len(action_type_list) > 1:
        action_type_list = action_type_list.split('|')
    # Single action type
    else:
        action_type_list = [action_type_list]


    for s in range(len(rew_list)):

        if not ((s == 0 and df.iloc[p]["rand_id"] in s1_failed) or (s == 1 and df.iloc[p]["rand_id"] in s2_failed) or (s == 2 and df.iloc[p]["rand_id"] in s3_failed)):
            state_list_0 = str(df.iloc[p][STATES_NAMES_LIST[s]]).split('|')
            # Save this state to list of all states
            all_states.append(list(map(float, state_list_0)))
            all_states_ids.append(df.iloc[p]["rand_id"])
            all_states_sessions.append(s)

            # Only save transition if attention checks have also been passed for next session
            if not (s == 0 and df.iloc[p]["rand_id"] in s2_failed) and not (s == 1 and df.iloc[p]["rand_id"] in s3_failed):
                
                pers_single = save_transition_single(df, pers_single, p, s, 
                                                     state_list_0, 
                                                     rew_list,
                                                     action_type_list)
    
    return all_states, all_states_ids, all_states_sessions, pers_single


if __name__ == "__main__":

    # Current directory
    curr_directory = os.path.abspath(os.path.dirname(__file__))

    # To save output (i.e., Cronbach's alpha values)
    output_file_name = "extract_RL_samples_output.txt"

    df = pd.read_excel(curr_directory + "/Data/conversational_sessions_anonym.xlsx")
    # Choose columns that we need
    df = df[["rand_id", "effort_list",
             "action_type_index_list", "state_0", 
             "state_1", "state_2", "state_3", "state_4", 
             "attention_check_list", 
             "attention_check_2_list"]]


    # We get user characteristics from the pre-questionnaire. 
    # We get the variables scaled to the interval [0, 1].
    df = get_similarity_variables_pre_questionnaire(df, output_file_name)
    
    # Get user characteristics from the Prolific profiles, scaled to the interval
    # [0, 1]
    df = get_char_prolific_profile(df, output_file_name)

    # Get TTM-stage for quitting smoking based on prescreening questionnaire
    # Requires that we already got the number of quit attempts in the last year
    # from the pre-questionnaire.
    df = get_char_prescreening_questionnaire(df)

    # Get involvement from post-questionnaire
    df = get_inv_post_questionnaire(df, output_file_name)


    # We need to remove data points from when people failed 2 attention checks
    # We have no people who failed 2 attention checks in session 4 or session 5
    # We convert the attention check lists to strings like '4|2|0|2', that is
    # why we look at index 2 for session 2 and index 4 for session 3.
    s1_failed = [df.loc[i, "rand_id"] for i in range(len(df)) if str(df.loc[i, "attention_check_list"])[0] in ["1", "2", "0"] and str(df.loc[i, "attention_check_2_list"])[0] in ["2", "3", "4"]]
    s2_failed = [df.loc[i, "rand_id"] for i in range(len(df)) if len(str(df.loc[i, "attention_check_list"])) > 1 and str(df.loc[i, "attention_check_list"])[2] in ["1", "2", "0"] and len(str(df.loc[i, "attention_check_2_list"])) > 1 and str(df.loc[i, "attention_check_2_list"])[2] in ["2", "3", "4"]]
    s3_failed = [df.loc[i, "rand_id"] for i in range(len(df)) if len(str(df.loc[i, "attention_check_list"])) > 3 and str(df.loc[i, "attention_check_list"])[4] in ["1", "2", "0"] and len(str(df.loc[i, "attention_check_2_list"])) > 3 and str(df.loc[i, "attention_check_2_list"])[4] in ["2", "3", "4"]]


    # We collect all valid states to compute means.
    # So, we also use the first state of people who never arrived in session 2.
    all_states = []
    all_states_ids = []
    all_states_sessions = []


    # Now we need to split the data by session
    pers_single = []
    for p in range(len(df)):

        # People who did at least 2 sessions (i.e., have at least 1 effort value)
        if not pd.isnull(df.iloc[p]["effort_list"]):

            all_states, all_states_ids, all_states_sessions, pers_single = save_data_person(df, 
                                                                                            p, 
                                                                                            s1_failed, 
                                                                                            s2_failed, 
                                                                                            s3_failed, 
                                                                                            all_states, 
                                                                                            all_states_ids, 
                                                                                            all_states_sessions, 
                                                                                            pers_single)


        # People for whom we only have data from session 1     
        # We still capture their first state
        else:

            # If the person did not fail 2 attention checks in session 1
            if not df.iloc[p]["rand_id"] in s1_failed:
                state_list = str(df.iloc[p]["state_0"]).split('|')
                # If we do not have info on all state features
                for missing_feat in range(len(state_list), NUM_STATE_FEATURES):
                    state_list.append(np.nan)
                all_states.append(list(map(float, state_list)))
                all_states_ids.append(df.iloc[p]["rand_id"])
                all_states_sessions.append(0)
  

    # Create dataframe
    df_single = pd.DataFrame(pers_single, columns = ["s0", "s1", 'a', 
                                                     'r',
                                                     'rand_id', 
                                                     "Pers_OE",
                                                     "Pers_Extraversion",
                                                     "Pers_ES", 
                                                     "Pers_Conscientiousness",
                                                     "Pers_Agreeableness",
                                                     "TTM_PA", "PA_Identity", 
                                                     "Involvement",
                                                     "Session",
                                                     "Quitting_Self_Identity",
                                                     "Smoker_Self_Identity",
                                                     "Nonsmoker_Self_Identity",
                                                     "Smoker_Group_Identity",
                                                     "Nonsmoker_Group_Identity",
                                                     "Quitting_Self_Efficacy",
                                                     "NFC",
                                                     "Quit_Before_24h",
                                                     "Num_Quit_24h_Last_Year",
                                                     "Time_Since_Last_24h_Quit_Attempt",
                                                     "Sitting_Hours_Weekday",
                                                     "Sitting_Hours_Weekendday",
                                                     "age_bin",
                                                     "Highest_Completed_Education_Level",
                                                     "Household_Size",
                                                     "Smoking_Frequency",
                                                     "Weekly_Exercise",
                                                     "Household_Income",
                                                     "Personal_Income",
                                                     "Quit_Smoking_Status",
                                                     "Smoking_Status",
                                                     "Socioeconomic_Status",
                                                     "Gender_Male", 
                                                     "Gender_Female",
                                                     "Gender identity",
                                                     "TTM_Smoking"])


    # We need to compute the feature means to get binary state features
    feat_means = np.mean(all_states, axis = 0)

    # We compute a binary version of the state features based on the feature means.
    # If we have partial data on a state for a session, we set the remaining 
    # feature values to 1 (i.e., like the mean) for that session.
    for row in range(len(df_single)):
        df_single.at[row, 's0'] = [0 if df_single.loc[row, 's0'][i] < feat_means[i] else 1 for i in range(NUM_STATE_FEATURES)]
        df_single.at[row, 's1'] = [0 if df_single.loc[row, 's1'][i] < feat_means[i] else 1 for i in range(NUM_STATE_FEATURES)]

    # Save as csv-file.
    df_single.to_csv(curr_directory + "/Analysis/Data/rl_samples_list_binary_activities.csv", 
                     index = False)


    ### Create a dataframe with all states and corresponding user IDs
    # and characteristics
    df_all_states = pd.DataFrame(all_states, columns = ["F0", "F1", "F2", "F3", 
                                                        "F4", "F5", "F6", "F7", 
                                                        "F8", "F9"])
    df_all_states["rand_id"] = all_states_ids
    df_all_states["Session"] = all_states_sessions

    # Get user characteristics
    df_all_states = pd.merge(df_all_states, df, on="rand_id", how='left')
    df_all_states = df_all_states.drop(columns=["effort_list",
                                                "attention_check_list", 
                                                "attention_check_2_list",
                                                "state_0", "state_1", "state_2", 
                                                "state_3", "state_4",
                                                "action_type_index_list"])

    # Get binary states
    binary_states = []
    for row in range(len(df_all_states)):
        binary_state_row = [1 if df_all_states.loc[row, 'F' + str(i)] >= feat_means[i] else 0 for i in range(NUM_STATE_FEATURES)]
        binary_states.append(binary_state_row)
    df_all_states["Binary_State"] = binary_states

    # Save data frame
    df_all_states.to_csv(curr_directory + "/Analysis/Data/all_states.csv", 
                         index = False)
